from ._fold_detector import FoldDetector
from ...process._processor import Processor

class RulesBasedFoldDetector(FoldDetector):
    def __init__(self, preprocessor: Processor = None, postprocessor: Processor = None) -> None:
        super().__init__("Rules Based Tissue Fold Detector", preprocessor, postprocessor)



from ._normaliser import Normaliser
from ..process._processor import Processor
import dask.array as da
from dask.array import Array
from skimage import img_as_ubyte
from skimage.color import rgb2lab, lab2rgb

class ReinhardNormaliser(Normaliser):
    def __init__(self, preprocessor: Processor = None, postprocessor: Processor = None) -> None:
        super().__init__("Reinhard Stain Normaliser", preprocessor, postprocessor)

    def fit(self, image: Array) -> None:
        assert isinstance(image, da.Array), "Image must be a dask array"
        assert image.ndim == 3, "Image must have 3 dimensions"
        assert image.shape[2] == 3, "Image must have 3 channels"
        image = self.preprocessor.process(image)
        image = da.map_blocks(rgb2lab, image, dtype="float32")
        self.means = da.mean(image, axis=(0, 1))
        self.stds = da.std(image, axis=(0, 1))
        self.fitted = True

    def run(self, image: Array) -> Array:
        assert self.fitted, "Normaliser must be fitted before running"
        assert isinstance(image, da.Array), "Image must be a dask array"
        assert image.ndim == 3, "Image must have 3 dimensions"
        assert image.shape[2] == 3, "Image must have 3 channels"
        image = self.preprocessor.process(image)
        image = da.map_blocks(rgb2lab, image, dtype="float32")
        means = da.mean(image, axis=(0, 1))
        stds = da.std(image, axis=(0, 1))
        image = (image - means) / stds * self.stds + self.means
        image = da.map_blocks(lambda x: img_as_ubyte(lab2rgb(x)), image, dtype="uint8")
        image = self.postprocessor.process(image)
        return image

    def fit_and_run(self, image: Array) -> Array:
        self.fit(image)
        return self.run(image)

    def _default_preprocessor(self) -> Processor:
        return Processor()

    def _default_postprocessor(self) -> Processor:
        return Processor()



from .._process import Process
import dask.array as da
from dask.array import Array
from skimage.color import rgb2lab, lab2rgb
from skimage import img_as_ubyte

class LuminosityPercentileStandardiser(Process):
    def __init__(self, percentile: float = 0.95) -> None:
        super().__init__("Standardise Luminosity by Percentile")
        assert isinstance(percentile, float), "Percentile must be a float"
        assert 0 < percentile < 1, "Percentile must be between 0 and 1"
        self.percentile = percentile

    def process(self, image: Array) -> Array:
        image = da.map_blocks(rgb2lab, image, dtype=float)
        luminosity = image[:, :, 0]
        luminosity = da.percentile(luminosity.ravel(), self.percentile)
        image[:, :, 0] = da.clip(image[:, :, 0], 0, luminosity)
        image = da.map_blocks(lab2rgb, image, dtype=float)
        image = da.map_blocks(img_as_ubyte, image, dtype="uint8")
        return image



from .._qc_step import QCStep
from ...process._processor import Processor
import dask.array as da
from dask.array import Array
from dask_image.ndmorph import binary_dilation
import numpy as np
from skimage.color import rgb2hsv

class FoldDetector(QCStep):
    def __init__(self, name: str, preprocessor: Processor = None, postprocessor: Processor = None) -> None:
        super().__init__(name, preprocessor, postprocessor)
        self.fitted = True

    def fit(self, image: Array) -> None:
        pass

    def run(self, image: Array) -> Array:
        image = da.map_blocks(rgb2hsv, image, dtype=float)
        delta = image[:, :, 1] - image[:, :, 2]
        delta_mask = da.greater(delta, 0)
        structure = np.array([
            [0, 0, 1, 0, 0],
            [0, 1, 1, 1, 0],
            [1, 1, 1, 1, 1],
            [0, 1, 1, 1, 0],
            [0, 0, 1, 0, 0]
        ])
        delta_mask_dilated = binary_dilation(delta_mask, structure=structure)
        delta_mask_lower = da.greater(delta, -0.3)
        delta_mask_dilated = da.logical_and(delta_mask_dilated, delta_mask_lower)
        return delta_mask_dilated

    def fit_and_run(self, image: Array) -> Array:
        return self.run(image)

    def _default_preprocessor(self) -> Processor:
        return Processor()

    def _default_postprocessor(self) -> Processor:
        return Processor()
